/*
 * UNICODE
 * $Id$
 *
 * Yet another Unicode string type for Python.  This type supports the
 * 16-bit Basic Multilingual Plane (BMP) only.
 *
 * Note that this string class supports embedded NULL characters.  End
 * of string is given by the length attribute.  However, the internal
 * representation always stores a trailing NULL to make it easier to
 * use unicode strings with standard APIs.
 *
 * History:
 * 1999-01-23 fl  Created
 * 1999-01-24 fl  Added split, join, capwords; basic UTF-8 support
 * 1999-01-24 fl  Basic UCS-2 support, buffer interface, etc.
 * 1999-03-06 fl  Moved declarations to separate file, etc.
 *
 * Written by Fredrik Lundh, January 1999.
 *
 * Copyright (c) 1999 by Secret Labs AB.
 * Copyright (c) 1999 by Fredrik Lundh.
 *
 * fredrik@pythonware.com
 * http://www.pythonware.com
 *
 * --------------------------------------------------------------------
 * This Unicode String Type is
 * 
 * Copyright (c) 1999 by Secret Labs AB
 * Copyright (c) 1999 by Fredrik Lundh
 * 
 * By obtaining, using, and/or copying this software and/or its
 * associated documentation, you agree that you have read, understood,
 * and will comply with the following terms and conditions:
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * associated documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appears in all
 * copies, and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Secret Labs
 * AB or the author not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.
 * 
 * SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS.  IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * -------------------------------------------------------------------- */

#include "Python.h"

#include "unicodeobject.h"

#if defined(HAVE_LIMITS_H)
#include <limits.h>
#else
#define INT_MAX 2147483647
#endif

/* -------------------------------------------------------------------- */
/* debugging stuff */

#if 0
static void
dump(PyUnicodeObject* self, const char* text)
{
    int i;
    printf("%p - ", self->str);
    if (text)
        printf("%s - ", text);
    for (i = 0; i < self->length; i++)
        printf("[%d]", self->str[i]);
    printf("\n");
}
#endif

static PyUnicodeObject* unicode_empty = NULL;

static void
unicode_dealloc(PyUnicodeObject* self)
{
    if (self->ref)
        Py_DECREF(self->ref);
    else
        free(self->str);
    PyMem_DEL(self);
}

PyObject*
PyUnicode_FromDataAndSize(const void* raw_content, int length)
{
    PyUnicodeObject* self;

    /* optimization for empty strings */
    if (length == 0 && unicode_empty != NULL) {
        Py_INCREF(unicode_empty);
        return (PyObject *)unicode_empty;
    }

    self = PyObject_NEW(PyUnicodeObject, &PyUnicode_Type);
    if (self == NULL)
	return NULL;

    self->ref = NULL;
    self->str = malloc((length+1)*sizeof(UNICODE_T));
    if (!self->str) {
        /* no room for the buffer */
        unicode_dealloc(self);
        PyErr_NoMemory();
        return NULL;
    }

    if (raw_content)
        UNICODE_COPY(self->str, raw_content, length);

    /* always add a trailing nul character */
    self->str[length] = 0;

    self->length = length;

    return (PyObject *)self;
}

PyUnicodeObject*
PyUnicode_New(int length)
{
    /* FIXME: for now... */
    return (PyUnicodeObject *)PyUnicode_FromDataAndSize(NULL, length);
}

PyObject*
PyUnicode_FromStringAndSize(const char* s, int size)
{
    int i;
    PyUnicodeObject* out;
    UNICODE_T* p;

    out = PyUnicode_New(size);
    if (!out)
        return NULL;

    p = out->str;

    for (i = 0; i < size; i++)
        *p++ = (unsigned char) *s++;

    return (PyObject *)out;
}

PyObject*
PyUnicode_FromString(const char* s)
{
    return PyUnicode_FromStringAndSize(s, strlen(s));
}

/*const*/UNICODE_T *
PyUnicode_AsUnicode(PyObject *op)
{
	if (!PyUnicode_Check(op)) {
		PyErr_BadInternalCall();
		return NULL;
	}
	return ((PyUnicodeObject *)op) -> str;
}

int PyUnicode_Size(PyObject *op)
{
	if (!PyUnicode_Check(op)) {
		PyErr_BadInternalCall();
		return -1;
	}
	return ((PyUnicodeObject *)op) -> length;
}

static char utf8_code_length[256] = {
    /* map UTF-8 encoded prefix byte to sequence length.  zero means
       illegal prefix.  see RFC 2279 for details*/
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 0, 0
};

PyObject*
PyUnicode_FromUTF8AndSize(const char* s, int size)
{
    int n;
    char* e;
    PyUnicodeObject* out;
    UNICODE_T* p;
    UNICODE_T ch;

    out = PyUnicode_New(size);
    if (!out)
        return NULL;

    p = out->str;

    e = (char*) s + size;

    /* unpack UTF-8 encoded data */
    while (s < e) {

        ch = (unsigned char) *s;

        if (ch < 0x80) {
            *p++ = ch;
            s++;
            continue;
        }

        n = utf8_code_length[ch];

        if (s + n > e) {
            PyErr_SetString(
                PyExc_ValueError, "unexpected end of UTF-8 data"
                );
            goto abort;
        }

        switch (n) {
        case 0:
            PyErr_SetString(
                PyExc_ValueError, "unexpected UTF-8 code byte"
                );
            goto abort;
        case 2:
            if ((s[1]&0xc0) != 0x80)
                goto bogus; /* invalid UTF-8 data */
            ch = ((s[0]&0x1f)<<6) + (s[1]&0x3f);
            if (ch < 0x80)
                goto bogus; /* illegal UTF-8 encoding */
            break;
        case 3:
            if ((s[1]&0xc0) != 0x80 || (s[2]&0xc0) != 0x80)
                goto bogus; /* invalid UTF-8 data */
            ch = ((s[0]&0x0f)<<12) + ((s[1]&0x3f)<<6) + (s[2]&0x3f);
            if (ch < 0x800 || ch >= 0xd800 && ch < 0xe000)
                goto bogus; /* illegal UTF-8 encoding */
            break;
        default:
            /* other sizes only needed for UCS-4 */
            PyErr_SetString(
                PyExc_ValueError, "unsupported UTF-8 code size"
                );
            Py_DECREF(out);
            return NULL;
        }
        *p++ = ch;
        s += n;
    }

    /* adjust length (FIXME: should reclaim storage space) */
    out->length = p - out->str;

    return (PyObject *)out;

bogus:
    PyErr_SetString(
        PyExc_ValueError, "invalid UTF-8 code"
        );
abort:
    Py_DECREF(out);
    return NULL;
}

PyObject*
PyUnicode_FromUTF8(const char* s)
{
    return PyUnicode_FromUTF8AndSize(s, strlen(s));
}

static PyUnicodeObject*
make_unicode_shared(PyUnicodeObject* ref, int start, int end)
{
    PyUnicodeObject* self;

    /* FIXME: if this function is to be exported, we should change ref
       to a PyObject*, and check the type in here */

    self = PyObject_NEW(PyUnicodeObject, &PyUnicode_Type);
    if (self == NULL)
	return NULL;

    /* find the first non-shared unicode object */
    while (ref->ref)
        ref = (PyUnicodeObject*) ref->ref;

    Py_INCREF(ref);

    self->ref = (PyObject*) ref;
    self->str = ref->str + start;
    self->length = end - start;

    return self;
}

static PyObject*
unicode_new(PyObject* self, PyObject* args)
{
    /* initialize unicode string from string buffer */

    /* by default, the string buffer is expected to contain plain US
       ASCII or UTF-8.  you can also pass in Latin-1 data by setting
       the format argument to latin-1 or iso-8859-1.  future versions
       will support more formats, via stream converters. */

    char* s;
    int size;
    char* fmt = "utf-8";
    if (!PyArg_ParseTuple(args, "s#|s", &s, &size, &fmt))
        return NULL;

    /* FIXME: yes, this is a hack... */
    if (strcmp(fmt, "us-ascii") == 0 || strcmp(fmt, "latin-1") == 0 ||
        strcmp(fmt, "iso-8859-1") == 0)
        return (PyObject*) PyUnicode_FromStringAndSize(s, size);

    if (strcmp(fmt, "utf-8") == 0)
        return (PyObject*) PyUnicode_FromUTF8AndSize(s, size);

    PyErr_SetString(
        PyExc_ValueError, "unsupported format"
        );

    return NULL;
}

static PyObject*
unicode_fromstring(PyObject* self, PyObject* args)
{
    PyUnicodeObject* out;

    /* FIXME: this one sucks.  get rid of it... */

    char* content;
    int size;
    int byteswap = 0;
    if (!PyArg_ParseTuple(args, "s#|i", &content, &size, &byteswap))
        return NULL;

    if (size & 1) {
        PyErr_BadArgument();
        return NULL;
    }

    out = (PyUnicodeObject *)PyUnicode_FromDataAndSize(content, size/2);

    if (out && byteswap) {
        /* swap bytes in place (???) */
        int len = out->length;
        UNICODE_T* s = out->str;
        while (len-- > 0)
            *s++ = ((*s >> 8) & 255) + (*s << 8);
    }

    return (PyObject*) out;
}

/* -------------------------------------------------------------------- */
/* method helpers */

static int
count(PyUnicodeObject* self, int start, int end, PyUnicodeObject* substring)
{
    int count = 0;

    end -= substring->length;

    while (start <= end)
        if (UNICODE_MATCH(self, start, substring)) {
            count++;
            start += substring->length;
        } else
            start++;

    return count;
}

static PyObject *
findstring(PyUnicodeObject* self, PyObject* args, int direction)
{
    PyUnicodeObject* substring;
    int start = 0;
    int end = INT_MAX;
    if (!PyArg_ParseTuple(args, "O!|ii", &PyUnicode_Type, &substring,
                          &start, &end))
        return NULL;

    if (start < 0)
        start += self->length;
    if (start < 0)
        start = 0;

    if (substring->length == 0)
        return PyInt_FromLong((long) start);

    if (end > self->length)
        end = self->length;
    if (end < 0)
        end += self->length;
    if (end < 0)
        end = 0;

    end -= substring->length;

    if (direction < 0) {
        for (; end >= start; end--)
            if (UNICODE_MATCH(self, end, substring))
                return PyInt_FromLong((long) end);
    } else {
        for (; start <= end; start++)
            if (UNICODE_MATCH(self, start, substring))
                return PyInt_FromLong((long) start);
    }

    return PyInt_FromLong(-1L);
}

static UNICODE_T*
findchar(PyUnicodeObject* self, UNICODE_T ch)
{
    /* like wcschr, but doesn't stop at NULL characters */

    UNICODE_T* s = self->str;
    int len = self->length;

    while (len-- > 0) {
        if (*s == ch)
            return s;
        s++;
    }

    return NULL;
}

static PyUnicodeObject*
fixup(PyUnicodeObject* self, int (*fixup)(PyUnicodeObject* s))
{
    /* apply fixup filter to unicode object */

    PyUnicodeObject* out;

    out = (PyUnicodeObject *)PyUnicode_FromDataAndSize(self->str, self->length);
    if (out && fixup)
        if (!fixup(out)) {
            /* callback should return TRUE if it modified the
               buffer. if FALSE, return a reference to the original
               buffer instead (to save space, not time) */
            Py_INCREF(self);
            Py_DECREF(out);
            return self;
        }
    return out;
}

static int
fixupper(PyUnicodeObject* self)
{
    int len = self->length;
    UNICODE_T* s = self->str;
    int status = 0;
    
    while (len-- > 0) {
        if (UNICODE_ISLOWER(*s)) {
            *s = UNICODE_TOUPPER(*s);
            status = 1;
        }
        s++;
    }

    return status;
}

static int
fixlower(PyUnicodeObject* self)
{
    int len = self->length;
    UNICODE_T* s = self->str;
    int status = 0;
    
    while (len-- > 0) {
        if (UNICODE_ISUPPER(*s)) {
            *s = UNICODE_TOLOWER(*s);
            status = 1;
        }
        s++;
    }

    return status;
}

static int
fixswapcase(PyUnicodeObject* self)
{
    int len = self->length;
    UNICODE_T* s = self->str;
    int status = 0;
    
    while (len-- > 0) {
        if (UNICODE_ISUPPER(*s)) {
            *s = UNICODE_TOLOWER(*s);
            status = 1;
        } else if (UNICODE_ISLOWER(*s)) {
            *s = UNICODE_TOUPPER(*s);
            status = 1;
        }
        s++;
    }

    return status;
}

static int
fixcapitalize(PyUnicodeObject* self)
{
    if (self->length > 0 && UNICODE_ISLOWER(self->str[0])) {
            self->str[0] = UNICODE_TOUPPER(self->str[0]);
            return 1;
    }

    return 0;
}

static PyObject*
join(PyObject* data, PyUnicodeObject* separator)
{
    int i, j, n;
    PyUnicodeObject* item;
    PyUnicodeObject* out;
    int sep_len;
    UNICODE_T* sep;
    UNICODE_T blank;

    if (!PySequence_Check(data)) {
	PyErr_SetString(
            PyExc_TypeError,
            "first argument must be a sequence"
            );
	return NULL;
    }

    if (separator) {
        sep = separator->str;
        sep_len = separator->length;
    } else {
        blank = ' ';
        sep = &blank;
        sep_len = 1;
    }

    n = PyObject_Length(data);
    if (n == 0)
        return (PyObject*) PyUnicode_New(0); /* empty list */

    /* first pass: check types, and determine size of output string */

    for (i = j = 0; i < n; i++) {
        if ((item = PyUnicode_Check(PySequence_GetItem(data, i))) == NULL) {
            PyErr_SetString(
                PyExc_TypeError,
                "first argument must be sequence of unicode strings"
                );
            return NULL;
        }
        j += item->length;
        Py_DECREF(item);
    }

    if (n == 1) {
        Py_INCREF(item); /* return first item */
        return (PyObject*) item;
    }

    /* second pass: create output string and fill it */

    out = PyUnicode_New(j + (n-1) * sep_len);
    if (!out)
        return NULL;

    for (i = j = 0; i < n-1; i++) {
        item = (PyUnicodeObject*) PySequence_GetItem(data, i);
        UNICODE_COPY(out->str + j, item->str, item->length);
        UNICODE_COPY(out->str + j + item->length, sep, sep_len);
        j += item->length + sep_len;
        Py_DECREF(item);
    }
    item = (PyUnicodeObject*) PySequence_GetItem(data, i);
    UNICODE_COPY(out->str + j, item->str, item->length);
    Py_DECREF(item);

    return (PyObject*) out;
}

static PyUnicodeObject *
pad(PyUnicodeObject* self, int left, int right, UNICODE_T fill)
{
    PyUnicodeObject* out;

    if (left < 0)
        left = 0;
    if (right < 0)
        right = 0;

    if (left == 0 && right == 0) {
        Py_INCREF(self);
        return self;
    }

    out = PyUnicode_New(left + self->length + right);
    if (out) {
        if (left)
            UNICODE_FILL(out->str, fill, left);
        UNICODE_COPY(out->str + left, self->str, self->length);
        if (right)
            UNICODE_FILL(out->str + left + self->length, fill, right);
    }

    return out;
}

static PyObject*
split(PyUnicodeObject* self, PyUnicodeObject* substring, int maxcount)
{
    PyObject* out;
    PyUnicodeObject* str;
    int i, j, status;

    /* NOTE: this function returns a list of references instead of
       copies.  it's definitely better for some use cases, but much
       worse for others... */

    if (!maxcount)
        maxcount = INT_MAX;

    out = PyList_New(0);
    if (!out)
        return NULL;

#define SPLIT_APPEND(self, start, end)\
    str = make_unicode_shared(self, (start), (end));\
    if (!str) goto abort;\
    status = PyList_Append(out, (PyObject*) str);\
    Py_DECREF(str);\
    if (!status < 0) goto abort;

    if (substring == NULL) {

        /* split on whitespace */
        for (i = 0; i < self->length;) {
            /* find a token */
            while (i < self->length && UNICODE_ISSPACE(self->str[i]))
                i++;
            j = i;
            while (i < self->length && !UNICODE_ISSPACE(self->str[i]))
                i++;
            if (j < i) {
                SPLIT_APPEND(self, j, i);
                while (i < self->length && UNICODE_ISSPACE(self->str[i]))
                    i++;
                j = i;
                if (maxcount-- <= 0)
                    break;
            }
        }
        if (j < self->length) {
            SPLIT_APPEND(self, j, self->length);
        }

    } else if (substring->length == 1) {

        /* split on single character */
        for (i = j = 0; i < self->length;)
            if (self->str[i] == substring->str[0]) {
                SPLIT_APPEND(self, j, i);
                i = j = i + 1;
                if (maxcount-- <= 0)
                    break;
            } else
                i++;
        if (j <= self->length) {
            SPLIT_APPEND(self, j, self->length);
        }

    } else {

        /* split on substring */
        for (i = j = 0; i < self->length - substring->length;)
            if (UNICODE_MATCH(self, i, substring)) {
                SPLIT_APPEND(self, j, i);
                i = j = i + substring->length;
                if (maxcount-- <= 0)
                    break;
            } else
                i++;
        if (j <= self->length) {
            SPLIT_APPEND(self, j, self->length);
        }
    }

    return (PyObject*) out;

abort:
    Py_DECREF(out);
    return NULL;

#undef SPLIT_APPEND

}

static PyObject *
strip(PyUnicodeObject* self, int left, int right)
{
    UNICODE_T* p = self->str;
    int start = 0;
    int end = self->length;

    if (left)
        while (start < end && UNICODE_ISSPACE(p[start]))
            start++;

    if (right)
        while (end > start && UNICODE_ISSPACE(p[end-1]))
            end--;

    if (start == 0 && end == self->length) {
        /* couldn't strip anything off, return original string */
        Py_INCREF(self);
        return (PyObject *)self;
    }

    return PyUnicode_FromDataAndSize(self->str + start, end - start);
}

/* -------------------------------------------------------------------- */
/* methods (in alphabetical order) */

static PyObject*
unicode_capitalize(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) fixup(self, fixcapitalize);
}

static PyObject*
unicode_capwords(PyUnicodeObject* self, PyObject* args)
{
    PyObject* list;
    PyObject* item;
    int i;

    if (!PyArg_NoArgs(args))
        return NULL;

    list = split(self, NULL, 0);
    if (!list)
        return NULL;

    /* capitalize each word */
    for (i = 0; i < PyList_GET_SIZE(list); i++) {
        item = (PyObject*) fixup(
            (PyUnicodeObject*) PyList_GET_ITEM(list, i),
            fixcapitalize
            );
        if (!item)
            goto abort;
        Py_DECREF(PyList_GET_ITEM(list, i));
        PyList_GET_ITEM(list, i) = item;
    }
    
    item = join(list, NULL);

abort:
    Py_DECREF(list);
    return (PyObject*) item;
}

static PyObject *
unicode_center(PyUnicodeObject* self, PyObject* args)
{
    int marg, left;

    int width;
    if (!PyArg_ParseTuple(args, "i", &width))
        return NULL;

    if (self->length >= width) {
        Py_INCREF(self);
        return (PyObject*) self;
    }

    marg = width - self->length;
    left = marg / 2 + (marg & width & 1);

    return (PyObject*) pad(self, left, marg - left, ' ');
}

static int
unicode_compare(PyUnicodeObject* str1, PyUnicodeObject* str2)
{
    int len1, len2;
    UNICODE_T *s1 = str1->str;
    UNICODE_T *s2 = str2->str;

    len1 = str1->length;
    len2 = str2->length;

    while (len1 > 0 && len2 > 0) {
        int cmp = (*s1++) - (*s2++);
        if (cmp)
            /* thist should make christian happy! */
            return (cmp < 0) ? -1 : (cmp != 0);
        len1--, len2--;
    }

    return (len1 < len2) ? -1 : (len1 != len2);
}

static PyObject*
unicode_concat(PyUnicodeObject* str1, PyObject* str2_)
{
    PyUnicodeObject* str2;
    PyUnicodeObject* out;

    if ((str2 = PyUnicode_Check(str2_)) == NULL) {
        PyErr_BadArgument();
        return NULL;
    }

    if (str2->length == 0) {
        Py_INCREF(str1);
        return (PyObject*) str1;
    }
    if (str1->length == 0) {
        Py_INCREF(str2);
        return (PyObject*) str2;
    }

    out = PyUnicode_New(str1->length + str2->length);
    if (!out)
        return NULL;

    UNICODE_COPY(out->str, str1->str, str1->length);
    UNICODE_COPY(out->str + str1->length, str2->str, str2->length);

    return (PyObject*) out;
}

static PyObject *
unicode_count(PyUnicodeObject* self, PyObject* args)
{
    PyUnicodeObject* substring;
    int start = 0;
    int end = INT_MAX;
    if (!PyArg_ParseTuple(args, "O!|ii", &PyUnicode_Type, &substring,
                          &start, &end))
        return NULL;

    if (substring->length == 0)
        return PyInt_FromLong((long) 0);

    if (start < 0)
        start += self->length;
    if (start < 0)
        start = 0;
    if (end > self->length)
        end = self->length;
    if (end < 0)
        end += self->length;
    if (end < 0)
        end = 0;

    return PyInt_FromLong((long) count(self, start, end, substring));
}

static PyObject*
unicode_expandtabs(PyUnicodeObject* self, PyObject* args)
{
    UNICODE_T* e;
    UNICODE_T* p;
    UNICODE_T* q;
    int i, j;
    PyUnicodeObject* out;

    int tabsize = 8;
    if (!PyArg_ParseTuple(args, "|i", &tabsize))
	return NULL;

    /* first pass: determine size of output string */
    i = j = 0;
    e = self->str + self->length;
    for (p = self->str; p < e; p++)
        if (*p == '\t')
            j += tabsize - (j%tabsize);
        else {
            j++;
            if (*p == '\n') {
                i += j;
                j = 0;
            }
        }

    /* second pass: create output string and fill it */
    out = PyUnicode_New(i + j);
    if (!out)
        return NULL;

    i = 0;
    q = out->str;

    for (p = self->str; p < e; p++)
        if (*p == '\t') {
            j = tabsize - (i%tabsize);
            i += j;
            while (j-- > 0)
                *q++ = ' ';
        } else {
            *q++ = *p;
            i++;
            if (*p == '\n')
                i = 0;
        }

    return (PyObject*) out;
}

static PyObject *
unicode_find(PyUnicodeObject* self, PyObject* args)
{
    return findstring(self, args, 1);
}

static PyObject *
unicode_getitem(PyUnicodeObject* self, int index)
{
    if (index < 0 || index >= self->length) {
        PyErr_SetString(PyExc_IndexError, "string index out of range");
        return NULL;
    }

    return (PyObject*) PyUnicode_FromDataAndSize(&self->str[index], 1);
}

static long
unicode_hash(PyUnicodeObject* self)
{
    int len;
    unsigned char *p;
    long x;

    /* from stringobject.c: this simply treats the unicode string as
       a string of raw bytes.  in other words, hash(unicode) equals
       hash(unicode.tostring()). */

    len = self->length * sizeof(UNICODE_T);
    p = (unsigned char *) self->str;
    x = (*p << 7);
    while (--len >= 0)
        x = (1000003L*x) ^ *p++;
    x ^= self->length * sizeof(UNICODE_T);
    if (x == -1)
        x = -2;

    return x;
}

static PyObject *
unicode_index(PyUnicodeObject* self, PyObject* args)
{
    PyObject* result;
    
    result = findstring(self, args, 1);

    if (result && PyInt_AsLong(result) < 0) {
        Py_DECREF(result);
        PyErr_SetString(PyExc_ValueError, "substring not found");
        return NULL;
    }

    return result;
}

static PyObject*
unicode_join(PyUnicodeObject* self, PyObject* args)
{
    PyObject* data;
    PyUnicodeObject* separator = NULL;
    if (!PyArg_ParseTuple(args, "O|O!", &data, &PyUnicode_Type, &separator))
        return NULL;

    return join(data, separator);
}

static int
unicode_length(PyUnicodeObject* self)
{
    return self->length;
}

static PyObject *
unicode_ljust(PyUnicodeObject* self, PyObject* args)
{
    int width;
    if (!PyArg_ParseTuple(args, "i", &width))
        return NULL;

    if (self->length >= width) {
        Py_INCREF(self);
        return (PyObject*) self;
    }

    return (PyObject*) pad(self, 0, width - self->length, ' ');
}

static PyObject*
unicode_lower(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) fixup(self, fixlower);
}

static PyObject *
unicode_lstrip(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) strip(self, 1, 0);
}

static PyObject*
unicode_repeat(PyUnicodeObject* str, int len)
{
    PyUnicodeObject* out;
    UNICODE_T* p;

    if (len < 0)
        len = 0;

    if (len == 1) {
        /* no repeat, return original string */
        Py_INCREF(str);
        return (PyObject*) str;
    }
        
    out = PyUnicode_New(len * str->length);
    if (!out)
        return NULL;

    p = out->str;

    while (len-- > 0) {
        UNICODE_COPY(p, str->str, str->length);
        p += str->length;
    }

    return (PyObject*) out;
}

static PyObject*
unicode_replace(PyUnicodeObject* self, PyObject* args)
{
    PyUnicodeObject* out;

    PyUnicodeObject* str1;
    PyUnicodeObject* str2;
    int maxcount = INT_MAX;
    if (!PyArg_ParseTuple(args, "O!O!|i", &PyUnicode_Type, &str1,
                          &PyUnicode_Type, &str2, &maxcount))
        return NULL;

    if (str1->length == 1 && str2->length == 1) {
        int i;

        /* replace characters */
        if (!findchar(self, str1->str[0])) {
            /* nothing to replace, return original string */
            Py_INCREF(self);
            out = self;
        } else {
            out = (PyUnicodeObject *)PyUnicode_FromDataAndSize(self->str, self->length);
            if (out)
                for (i = 0; i < out->length; i++)
                    if (out->str[i] == str1->str[0]) {
                        if (maxcount-- <= 0)
                            break;
                        out->str[i] = str2->str[0];
                    }
        }

    } else {
        int n, i;
        UNICODE_T* p;

        /* replace strings */
        n = count(self, 0, self->length, str1);
        if (n > maxcount)
            n = maxcount;
        if (n == 0) {
            /* nothing to replace, return original string */
            Py_INCREF(self);
            out = self;
        } else {
            out = PyUnicode_New(
                self->length + n * (str2->length - str1->length)
                );
            if (out) {
                i = 0;
                p = out->str;
                while (i <= self->length - str1->length)
                    if (UNICODE_MATCH(self, i, str1)) {
                        /* replace string segment */
                        UNICODE_COPY(p, str2->str, str2->length);
                        p += str2->length;
                        i += str1->length;
                        if (--n <= 0) {
                            /* copy remaining part */
                            UNICODE_COPY(p, self->str+i, self->length-i);
                            break;
                        }
                    } else
                        *p++ = self->str[i++];
            }
        }
    }
    
    return (PyObject*) out;
}

static PyObject *
unicode_repr(PyUnicodeObject* self)
{
    PyObject* repr;
    UNICODE_T* s;
    char* p;
    char* q;
    int len;

    static const char* hexdigit = "0123456789abcdef";

    repr = PyString_FromStringAndSize(NULL, 1 + 6*self->length + 1);
    if (!repr)
        return NULL;

    p = q = PyString_AS_STRING(repr);
    s = self->str;
    len = self->length;

    *p++ = (findchar(self, '\'') && !findchar(self, '"')) ? '"' : '\'';
    while (len-- > 0) {
        UNICODE_T ch = *s++;
        if (ch == *q || ch == '\\') {
            *p++ = '\\';
            *p++ = (char) ch;
        } else if (ch >= 256) {
            /* map 16-bit characters to '\uxxxx' */
            *p++ = '\\';
            *p++ = 'u';
            *p++ = hexdigit[(ch >> 12) & 15];
            *p++ = hexdigit[(ch >> 8) & 15];
            *p++ = hexdigit[(ch >> 4) & 15];
            *p++ = hexdigit[ch & 15];
        } else if (ch < ' ' || ch >= 128) {
            /* map non-printable US ASCII to '\ooo' */
            *p++ = '\\';
            *p++ = hexdigit[(ch >> 6) & 7];
            *p++ = hexdigit[(ch >> 3) & 7];
            *p++ = hexdigit[ch & 7];
        } else
            *p++ = (char) ch;
    }
    *p++ = q[0];

    *p = '\0';

    _PyString_Resize(&repr, p - q);

    return repr;
}

static PyObject *
unicode_rfind(PyUnicodeObject* self, PyObject* args)
{
    return findstring(self, args, -1);
}

static PyObject *
unicode_rindex(PyUnicodeObject* self, PyObject* args)
{
    PyObject* result;

    result = findstring(self, args, -1);

    if (result && PyInt_AsLong(result) < 0) {
        Py_DECREF(result);
        PyErr_SetString(PyExc_ValueError, "substring not found");
        return NULL;
    }

    return result;
}

static PyObject *
unicode_rjust(PyUnicodeObject* self, PyObject* args)
{
    int width;
    if (!PyArg_ParseTuple(args, "i", &width))
        return NULL;

    if (self->length >= width) {
        Py_INCREF(self);
        return (PyObject*) self;
    }

    return (PyObject*) pad(self, width - self->length, 0, ' ');
}

static PyObject *
unicode_rstrip(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) strip(self, 0, 1);
}

static PyObject*
unicode_slice(PyUnicodeObject* self, int start, int end)
{
    /* standard clamping */
    if (start < 0)
        start = 0;
    if (end < 0)
        end = 0;
    if (end > self->length)
        end = self->length;
    if (start == 0 && end == self->length) {
        /* full slice, return original string */
        Py_INCREF(self);
        return (PyObject*) self;
    }
    if (start > end)
        start = end;
    /* copy slice */
    return (PyObject*) PyUnicode_FromDataAndSize(
        self->str + start,
        end - start);
}

static PyObject*
unicode_slice_shared(PyUnicodeObject* self, PyObject* args)
{
    /* experimental slicing.  this keeps a reference to the original
       unicode object.  use with care. */

    int start = 0;
    int end = INT_MAX;
    if (!PyArg_ParseTuple(args, "|ii", &start, &end))
        return NULL;

    /* standard clamping */
    if (start < 0)
        start = 0;
    if (end < 0)
        end = 0;
    if (end > self->length)
        end = self->length;
    if (start == 0 && end == self->length) {
        /* full slice, return original string */
        Py_INCREF(self);
        return (PyObject*) self;
    }
    if (start > end)
        start = end;
    /* return reference to original string */
    return (PyObject*) make_unicode_shared(self, start, end);
}

static PyObject*
unicode_split(PyUnicodeObject* self, PyObject* args)
{
    PyUnicodeObject* substring = NULL;
    int maxcount = 0;
    if (!PyArg_ParseTuple(args, "|O!i", &PyUnicode_Type, &substring,
                          &maxcount))
        return NULL;

    return split(self, substring, maxcount);
}

PyObject *
PyUnicode_ToUTF8(PyObject* pyself)
{
	PyUnicodeObject *self;
    PyObject* str;
    UNICODE_T* s;
    char* p;
    char* q;
    int len;

	if (!PyUnicode_Check(pyself)) {
		PyErr_SetString(PyExc_TypeError, "Object is not a PyUnicode");
		return NULL;
	}
	self = (PyUnicodeObject *)pyself;

    str = PyString_FromStringAndSize(NULL, 3 * self->length);
    if (!str)
        return NULL;

    p = q = PyString_AS_STRING(str);
    s = self->str;
    len = self->length;

    if (len > 0) {
        
        /* generate UTF-8 representation */

        while (len-- > 0) {
            UNICODE_T ch = *s++;
            if (ch < 0x80)
                *p++ = (char) ch;
            else {
                if (ch < 0x800) {
                    *p++ = 0xc0 | (ch >> 6);
                    *p++ = 0x80 | (ch & 0x3f);
                } else {
                    *p++ = 0xe0 | (ch >> 12);
                    *p++ = 0x80 | ((ch >> 6) & 0x3f);
                    *p++ = 0x80 | (ch & 0x3f);
                }
            }
        }

        *p = '\0';

        _PyString_Resize(&str, p - q);
        
    }

    return str;
}

static PyObject *
unicode_strip(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) strip(self, 1, 1);
}

static PyObject*
unicode_swapcase(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) fixup(self, fixswapcase);
}

static PyObject *
unicode_tostring(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return PyString_FromStringAndSize(
        (char*) self->str, self->length*sizeof(UNICODE_T)
        );
}

static PyObject*
unicode_upper(PyUnicodeObject* self, PyObject* args)
{
    if (!PyArg_NoArgs(args))
        return NULL;
    return (PyObject*) fixup(self, fixupper);
}

static PyObject *
unicode_zfill(PyUnicodeObject* self, PyObject* args)
{
    int fill;
    PyUnicodeObject* out;

    int width;
    if (!PyArg_ParseTuple(args, "i", &width))
        return NULL;

    if (self->length >= width) {
        Py_INCREF(self);
        return (PyObject*) self;
    }

    fill = width - self->length;

    out = pad(self, fill, 0, '0');

    if (out->str[fill] == '+' || out->str[fill] == '-') {
        /* move sign to beginning of string */
        out->str[0] = out->str[fill];
        out->str[fill] = '0';
    }

    return (PyObject*) out;
}

static PyMethodDef unicode_methods[] = {

    {"capitalize", (PyCFunction) unicode_capitalize, 0},
    {"capwords", (PyCFunction) unicode_capwords, 0},
    {"center", (PyCFunction) unicode_center, 1},
    {"count", (PyCFunction) unicode_count, 1},
    {"expandtabs", (PyCFunction) unicode_expandtabs, 1},
    {"find", (PyCFunction) unicode_find, 1},
    {"index", (PyCFunction) unicode_index, 1},
    {"join", (PyCFunction) unicode_join, 1},
    {"ljust", (PyCFunction) unicode_ljust, 1},
    {"lower", (PyCFunction) unicode_lower, 0},
    {"lstrip", (PyCFunction) unicode_lstrip, 0},
/*  {"maketrans", (PyCFunction) unicode_maketrans, 1}, */
    {"replace", (PyCFunction) unicode_replace, 1},
    {"rfind", (PyCFunction) unicode_rfind, 1},
    {"rindex", (PyCFunction) unicode_rindex, 1},
    {"rjust", (PyCFunction) unicode_rjust, 1},
    {"rstrip", (PyCFunction) unicode_rstrip, 0},
    {"slice", (PyCFunction) unicode_slice_shared, 1},
    {"split", (PyCFunction) unicode_split, 1},
    {"strip", (PyCFunction) unicode_strip, 0},
    {"swapcase", (PyCFunction) unicode_swapcase, 0},
    {"tostring", (PyCFunction) unicode_tostring, 0},
/*  {"translate", (PyCFunction) unicode_translate, 1}, */
    {"upper", (PyCFunction) unicode_upper, 0},
    {"zfill", (PyCFunction) unicode_zfill, 1},

    {NULL, NULL}
};

static PyObject*  
unicode_getattr(PyUnicodeObject* self, char* name)
{
    return Py_FindMethod(unicode_methods, (PyObject*) self, name);
}

static PySequenceMethods unicode_as_sequence = {
    (inquiry) unicode_length, /* sq_length */
    (binaryfunc) unicode_concat, /* sq_concat */
    (intargfunc) unicode_repeat, /* sq_repeat */
    (intargfunc) unicode_getitem, /* sq_item */
    (intintargfunc) unicode_slice, /* sq_slice */
    0, /* sq_ass_item */
    0, /* sq_ass_slice */
};

static int
unicode_buffer_getreadbuf(PyUnicodeObject* self, int index, const void **ptr)
{
    if (index != 0) {
        PyErr_SetString(
            PyExc_SystemError,
            "accessing non-existent unicode segment"
            );
        return -1;
    }
    *ptr = (void *) self->str;
    return self->length * sizeof(UNICODE_T);
}

static int
unicode_buffer_getwritebuf(PyUnicodeObject* self, int index, const void **ptr)
{
    PyErr_SetString(
        PyExc_TypeError,
        "cannot use unicode as modifyable buffer"
        );
    return -1;
}

static int
unicode_buffer_getsegcount(PyUnicodeObject* self, int* lenp)
{
    if (lenp)
        *lenp = self->length * sizeof(UNICODE_T);
    return 1;
}

static PyBufferProcs unicode_as_buffer = {
    (getreadbufferproc) unicode_buffer_getreadbuf,
    (getwritebufferproc) unicode_buffer_getwritebuf,
    (getsegcountproc) unicode_buffer_getsegcount
};

PyTypeObject PyUnicode_Type = {
    PyObject_HEAD_INIT(NULL)
    0, /* ob_size */
    "unicode", /* tp_name */
    sizeof(PyUnicodeObject), /* tp_size */
    0, /* tp_itemsize */
    /* methods */
    (destructor)unicode_dealloc, /* tp_dealloc */
    0, /* tp_print */
    (getattrfunc)unicode_getattr, /* tp_getattr */
    0, /* tp_setattr */
    (cmpfunc) unicode_compare, /* tp_compare */
    (reprfunc) unicode_repr, /* tp_repr */
    0, /* tp_as_number */
    &unicode_as_sequence, /* tp_as_sequence */
    0, /* tp_as_mapping */
    (hashfunc) unicode_hash /* tp_hash*/,
    0, /*tp_call*/
    (reprfunc) PyUnicode_ToUTF8, /* tp_str */
    (getattrofunc) NULL, /* tp_getattro */
    (setattrofunc) NULL, /* tp_setattro */
    &unicode_as_buffer /* tp_as_buffer */
};

static PyMethodDef _functions[] = {

    {"unicode", (PyCFunction) unicode_new, 1},

    {"join", (PyCFunction) unicode_join, 1},

    {"fromstring", (PyCFunction) unicode_fromstring, 1},

    {NULL, NULL}
};

void
#ifdef WIN32
__declspec(dllexport)
#endif
initunicode()
{
    PyObject* m;
    PyObject* d;
    UNICODE_T mark = 0xfeff;

    /* Patch object type */
    PyUnicode_Type.ob_type = &PyType_Type;

    if (sizeof(UNICODE_T) != 2)
        fprintf(
            stderr,
            "*** CONFIGURATION ERROR: sizeof(UNICODE_T) != 2 bytes ***\n"
            );

    unicode_empty = PyUnicode_New(0);

    m = Py_InitModule("unicode", _functions);
    d = PyModule_GetDict(m);

    /* add byte order mark to module dictionary */
    PyDict_SetItemString(
        d, "mark",
        (PyObject*) PyUnicode_FromDataAndSize(&mark, 1));
}
